// *****************************************************************
//
// constraint_store.cc
//
// Part of Polyanna, Richard Gault, 2001
//
// Copyright (C) University of Oxford, 2001, 2002
//
// *****************************************************************

#include "constraint_store_headers.h"

using namespace std;



// *****************************************************************
// ***                  ConstraintStore class                    ***
// *****************************************************************

// Note: The solver itself lives in the file solver.cc




ConstraintStore::~ConstraintStore()
{
  while (constraints.begin() != constraints.end())
  {
    Constraint* c = *(constraints.begin());
    delete(c);
    constraints.erase(constraints.begin());
  }
}




// ****************




void ConstraintStore::add_constraint(Constraint* c)
{
  constraints.insert(c);
  
  if (c->need_level_notification())
    notifiable_constraints.insert(c);
}




// ****************




bool ConstraintStore::remove_constraint(Constraint* c, bool del)
{
  set<Constraint*>::iterator location = constraints.find(c);
  if (location == constraints.end())
    return true;
  
  constraints.erase(location);
  if (c->need_level_notification())
    notifiable_constraints.erase(notifiable_constraints.find(c));
  
  if (del)
    delete(c);
  
  return false;
}




// ****************




ConstraintStore::ConstraintStore(nat a, const set<int>& dom)
{
  arity = a;
  domain = dom;

  vector<set<int>::const_iterator> ti(a);
  vector<int> t(a);   // t is obtained from ti by following iterators pointwise

  if (dom.size() == 0)
    return;

  // We create each a-tuple of domain elements, and put them in our set.
  // This is slow, but only done rarely.
  for (nat i = 0; i < a; ++i)
    ti[i] = dom.begin();
  
  for(;;)
  {
    nat i;
    int j;
    bool reset = true;
    
    // (Set up t from ti, and add it to the map)
    for (i = 0; i < a; ++i)
      t[i] = *(ti[i]);
    
    const FunctionArguments* fa = FunctionArguments::tuple_to_function_arguments(t);
    constrained_fas.insert(fa);
    lex_ordering.push_back(fa);
    
    // (Find lexicographically next tuple.  We can't use nat here, because we
    // undeflow.)
    for (j = a - 1, reset = true; j >= 0 && reset; --j)
    {
      ++(ti[j]);
      if (ti[j] == dom.end())
	ti[j] = dom.begin();
      else
	reset = false;
    }
    if (reset) break;     // We've "clocked over" to the first tuple once again
  }
}




// **********************




void ConstraintStore::display(ostream& s) const
{
  for (CSI csi = constraints.begin(); csi != constraints.end(); ++csi)
    (*csi)->display(s);
}




// **********************




AssociativeConstraint* ConstraintStore::add_associative_constraint()
{
  if (arity != 2)
  {
    cerr << "Error: Attempt made to check for associativity of polymorphisms of arity\n";
    cerr << "       other than two.\n";
    throw ConsistencyError();
  }
  
  AssociativeConstraint* c = new(AssociativeConstraint)(arity, domain);
  add_constraint(c);
  
  return c;
}




// **********************




CommutativeConstraint* ConstraintStore::add_commutative_constraint()
{
  if (arity != 2)
  {
    cerr << "Error: Attempt made to check for commutativity of polymorphisms of arity\n";
    cerr << "       other than two.\n";
    throw ConsistencyError();
  }
  
  CommutativeConstraint* c = new(CommutativeConstraint);
  add_constraint(c);
  
  return c;
}




// **********************




IdempotentConstraint* ConstraintStore::add_idempotent_constraint()
{
  IdempotentConstraint* c = new(IdempotentConstraint)(arity);
  add_constraint(c);
  
  return c;
}




// **********************




ZeroOneRangeConstraint* ConstraintStore::add_zero_one_range_constraint()
{
  ZeroOneRangeConstraint* c = new(ZeroOneRangeConstraint);
  add_constraint(c);
  
  return c;
}




// **********************




NearUnanimityConstraint* ConstraintStore::add_near_unanimity_constraint()
{
  if (arity < 3)
  {
    cerr << "Error: Attempt made to check majority/near-unanimity functions of arity\n";
    cerr << "       less than three.\n";
    throw ConsistencyError();
  }
  
  NearUnanimityConstraint* c = new(NearUnanimityConstraint)(arity);
  add_constraint(c);
  
  return c;
}




// ***********************




void ConstraintStore::add_affine_constraint()
{
  if (arity != 3)
  {
    cerr << "Error: Attempt made to check for affine functions of arity something other\n";
    cerr << "       than three.\n";
    throw ConsistencyError();
  }
  
  add_constraint(new(Affine1Constraint)(arity, domain));
  add_constraint(new(Affine2Constraint));
  add_constraint(new(Affine3Constraint));
}




// **********************




ConservativeConstraint* ConstraintStore::add_conservative_constraint()
{
  ConservativeConstraint* c = new(ConservativeConstraint)(arity);
  add_constraint(c);
  
  return c;
}




// **********************




MaltsevConstraint* ConstraintStore::add_maltsev_constraint()
{
  if (arity != 3)
  {
    cerr << "Error: Attempt made to check for Mal'tsev functions of arity something other\n";
    cerr << "       than three.\n";
    throw ConsistencyError();
  }
  
  MaltsevConstraint* c = new(MaltsevConstraint);
  add_constraint(c);
  
  return c;
}




// **********************




void ConstraintStore::add_two_semilattice_constraint()
{
  if (arity != 2)
  {
    cerr << "Error: Attempt made to check for 2-semilattice functions of arity something other\n";
    cerr << "       than two.\n";
    throw ConsistencyError();
  }
  
  add_constraint(new(TwoSemilattice1Constraint)(arity, domain));
  add_constraint(new(CommutativeConstraint));
  add_constraint(new(IdempotentConstraint)(arity));
}




// **********************




SquashingConstraint* ConstraintStore::add_squashing_constraint()
{
  if (arity != 1)
  {
    cerr << "Error: Attempt made to check for squashing functions of arity something other\n";
    cerr << "       than one.\n";
    throw ConsistencyError();
  }
  
  SquashingConstraint* c = new(SquashingConstraint)(domain);
  add_constraint(c);
  
  return c;
}




// **********************




AutomorphismConstraint* ConstraintStore::add_automorphism_constraint
                               (const RelationStore& relstore,
				bool argument_perms, bool da_left,
				bool da_right, bool da_both,
				bool accuracy)
{
  AutomorphismConstraint* c = new(AutomorphismConstraint)
                                     (constrained_fas, domain, relstore,
				      argument_perms,
				      da_left, da_right, da_both,
				      accuracy);
  add_constraint(c);
  
  return c;
}




// **********************




EssentiallyNAryConstraint* ConstraintStore::add_essentiallynary_constraint()
{
  EssentiallyNAryConstraint* c = new(EssentiallyNAryConstraint)
                                       (constrained_fas, domain, arity);
  add_constraint(c);
  
  return c;
}




// **********************




PolymorphismConstraint* ConstraintStore::add_polymorphism_constraint
                               (const set<const FunctionArguments*>& fas)
{
  PolymorphismConstraint* c = new(PolymorphismConstraint)(fas, domain, arity);
  add_constraint(c);
  
  return c;
}




// **********************



void ConstraintStore::check_store_consistency()
{
  // First -- really paranoid, this -- check that all the FunctionArguments
  // we're going to assign to have the correct arity
  for (set<const FunctionArguments*>::const_iterator
	 fasi = constrained_fas.begin();
       fasi != constrained_fas.end();
       ++fasi)
  {
    if ((*fasi)->get_arity() != arity)
      throw SanityError();
  }

  // Now construct a set of all constraint types in the store, complaining
  // about duplicates which shouldn't be there.
  // Also, allow all the constraints in the store to perform their own
  // consistency checks.
  set<Constraint::ConstraintType> types;
  bool duplicated_warning = false;       // Only warn once about duplicates

  for (CSI csi = constraints.begin(); csi != constraints.end(); ++csi)
  {
    Constraint::ConstraintType this_type = (*csi)->get_constraint_type();
    if (!duplicated_warning &&
	types.find(this_type) != types.end() &&
	this_type != Constraint::Polymorphism)
    {
      cerr << "Warning: Some constraints have been added to the store twice\n";
      duplicated_warning = true;
    }
    
    types.insert(this_type);

    // Get the constraint to check consistency
    (*csi)->check_consistency(arity, domain, constrained_fas);
  }

  // Are we eliminating argument permutations?
  bool arg_perms_enabled = false;
  for (CSI csi = constraints.begin(); csi != constraints.end(); ++csi)
  {
    if ((*csi)->get_constraint_type() == Constraint::Automorphism &&
	((AutomorphismConstraint*)(*csi))->argument_permutations_enabled())
      arg_perms_enabled = true;
  }
      
  // A Permutation constraint is redundant if we're searching for
  // commutative functions
  if (arg_perms_enabled && types.find(Constraint::Commutative) != types.end())
  {
    cerr << "Warning: Permutation checking is enabled whilst searching for commutative\n";
    cerr << "         polymorphims.  This is redundant.\n";
  }

  // If we've got permutation constraint turned on, then we'd better make
  // sure that all constraints are closed under permutations.
  if (arg_perms_enabled)
  {
    for (CSI csi = constraints.begin(); csi != constraints.end(); ++csi)
    {
      if (!(*csi)->permutation_closed() &&
	  (*csi)->get_constraint_type() != Constraint::Automorphism)
      {
	cerr << "ERROR: \"Eliminate argument permutations\" is enabled whilst searching for\n";
	cerr << "       a polymorphism class which is not closed under argument permutations.\n";
	cerr << "       This means that some polymorphisms would not be found.\n";
	throw ConsistencyError();
      }
    }
  }
}
