// ***********************************************************************
//
// front_end_headers.h
//
// Part of Polyanna, Richard Gault, 2001
//
// Copyright (C) University of Oxford, 2001, 2002
//
// ***********************************************************************

#ifndef FRONT_END_HEADERS

#define FRONT_END_HEADERS

#include <string>
#include "misc_headers.h"
#include "constraint_store_headers.h"

// ************************************************************************
// ***                          UserOptions Class                       ***
// ************************************************************************

class UserOptions
{
 public:
  // Constructor.  Sets up default options
  UserOptions();
  
  // Read opts from command line.  Returns posution of 1st filename
  nat read_options(nat argc, char* argv[]);

  // Display a brief message describing the program
  void display_short_help() const;

  // A fuller message, including the command-line options
  void display_long_help() const;

  // Arity of polymorphims we're searching for
  nat get_fn_arity() const
    {return fn_arity;}

  // Remove duplicated permutations?
  bool get_perm_opts() const
    {return perm_opts;}

  // Should all functions be essentially n-ary?
  bool get_essentially_nary() const
    {return essentiallynary;}

  // Should injective unary polymorphisms be factored out?
  bool get_automorphisms() const
    {return automorphisms;}

  // Associative functions
  bool get_associative() const
    {return associative;}

  // Commutative function
  bool get_commutative() const
    {return commutative;}

  // Idempotent functions
  bool get_idempotent() const
    {return idempotent;}

  // Near-unanimity functions
  bool get_near_unanimity() const
    {return near_unanimity;}

  // Affine function
  bool get_affine() const
    {return affine;}

   // Conservative functions
  bool get_conservative() const
    {return conservative;}

  // 2-semilattices
  bool get_twosemilattice() const
    {return twosemilattice;}

  // Mal'tsev functions
  bool get_maltsev() const
    {return maltsev;}

  // Squashing (i.e., non-injective) functions
  bool get_squash() const
    {return squash;}

  // All functions must satisfy all non-symmetry-breaking constraints
  bool get_force_essential_constraints() const
    {return force_essential_constraints;}

  // All functions must satisfy all symmetry-breaking constraints
  bool get_force_all_symmetry_constraints() const
    {return force_all_symmetry_constraints;}

  // All functions output must be "complete": no compact representations.
  bool get_force_complete_functions() const
    {return force_complete_functions;}

  // The force level (0-2), as an integer.
  nat get_force_level() const;

  // Stop when we find the first function
  bool get_stop_at_first() const
    {return stop_at_first;}

  // Don't display output
  bool get_quiet() const
    {return quiet;}

  // Display tally of solution specs we found
  bool get_count() const
    {return count;}

  // Do we just want to know if the relations are tractable?
  bool get_tractable() const
    {return tractable;}

  // Do we want to display the constraints?
  bool get_view_constraints() const
    {return view_constraints;}

  // Do we want to view the decomposed relations?
  bool get_view_relations() const
    {return view_relations;}

  // Perform the same calculation repeatedly.  Debugging only.
  nat get_iterations() const
    {return iterations;}

  // Domain of relations to enumerate.  -1 for "off".
  int get_enumerate_relations() const
    {return enumerate_relations;}

  // Display most important options (for post-title display)
  void display_options_summary() const;

  // ************
  //  Exceptions
  // ************

  // SanityError is used when an event which should never occur, occurs.
  class SanityError{};

  
  // An unknown command-line option has been passed
  class UnknownOption
    {
    public:
      UnknownOption(std::string s)
	{option = s;}
      std::string option;
    };

  
  // A command-line option has an incorrect syntax
  class MalformedOption
    {
    public:
      MalformedOption(std::string s)
	{option = s;}
      std::string option;
    };

  
 private:
  nat fn_arity;
  bool perm_opts;
  bool essentiallynary;
  bool automorphisms;
  bool associative;
  bool commutative;
  bool idempotent;
  bool affine;
  bool near_unanimity;
  bool conservative;
  bool twosemilattice;
  bool maltsev;
  bool squash;
  bool force_essential_constraints;
  bool force_all_symmetry_constraints;
  bool force_complete_functions;
  nat force_level;
  bool stop_at_first;
  bool quiet;
  bool count;
  bool tractable;
  bool view_constraints;
  bool view_relations;
  nat iterations;
  int enumerate_relations;
};







// **************************************************************







// **************************************************************
// ***              RelationEnumerator Class                  ***
// **************************************************************

// The RelationEnumerator class is just an objectified interface
// to a function which displays all essentially different relations
// of a given arity and over a domain of a given size.

class RelationEnumerator : public SolutionReceiver
{
 public:
  // Constuctor.  Supply arity and domain size
  RelationEnumerator(nat arity, nat domain_size);

  // Generate the relations, and (if !quiet) display them
  nat generate_relations(bool quiet = false);
  
 private:

  // Private copy of supplied arity...
  nat arity;

   // ... and domain size
  nat domain_size;


  // Receive characteristic fn of relations for output
  void polymorphism_receiver(const Function& f) const;
};

#endif

