// *********************************************************************
//
// functions.cc
//
// Part of Polyanna, Richard Gault, 2001
//
// Copyright (C) University of Oxford, 2001, 2002
//
// *********************************************************************

#include "function_headers.h"
#include "misc_headers.h"
#include <cmath>
#include <algorithm>

using namespace std;




// *********************************************************************
// ***                  FunctionArguments class                      ***
// *********************************************************************

map<vector<int>,const FunctionArguments*>
                                       FunctionArguments::tuple_to_fn_args_map;




void FunctionArguments::FunctionArguments_constructor(const vector<int> &t)
{
  tuple = t;
  generate_permutations();
  initialise_constant_function_arguments();
  initialise_almost_constant_function_arguments();
}




// **********************************************************




const FunctionArguments* FunctionArguments::tuple_to_function_arguments
                                            (const vector<int> &t)
{
  map<vector<int>, const FunctionArguments*>::const_iterator mi;
  mi = tuple_to_fn_args_map.find(t);
  
  if (mi != tuple_to_fn_args_map.end())
    return mi->second;
  
  // Below is the only place where a FunctionArgument object can be created.
  // We invoke its "constructor" manually after the object is created.  This
  // is because functions such as generate_permutations(), which is called by
  // the "constructor", may cause another FunctionArgument object to be
  // created.  In turn, this may try to reference the first object.  If we
  // used a genuine constructor, then the first object wouldn't have finished
  // being built, so we wouldn't have an address for it, so the second
  // object would not realise that the first object was already in the process
  // of being constructed, and would try to construct it again.  An infinte
  // regression would occur.
  FunctionArguments* fa = new(FunctionArguments);

  // Must store address *before* invoking "constructor"
  tuple_to_fn_args_map[t] = fa;
  fa->FunctionArguments_constructor(t);
  fa->resize_functions_definable(max_functions_definable);
  
  return fa;
}




// *********************************************************




void FunctionArguments::generate_permutations()
{
  // Create a vector to hold  next permutation of <0,1,2,3,..., tuple.size()-1>
  vector<nat> order(tuple.size());
  for (nat i = 0; i < tuple.size(); ++i)
    order[i] = i;

  // Loop over all non-trivial permutations
  while (next_permutation(order.begin(), order.end()))
  {
    vector<int> next_tuple(tuple.size());
    for (nat i = 0; i < tuple.size(); ++i)
      next_tuple[i] = tuple[order[i]];
    
    permutations.push_back(tuple_to_function_arguments(next_tuple));
  }
  
  // Set up rev_perm.  The reverse permutation is always the last in the list.
  if (permutations.size() > 0)
    rev_perm = permutations[permutations.size() - 1];
  else
    rev_perm = this;
}




// *********************************************************




void FunctionArguments::initialise_constant_function_arguments()
{
  // Note that this function works for size 0 and size 1 tuples
  constant_function_arguments = true;
  
  for (nat i = 1; i < tuple.size(); i++)
  {
    if (tuple[i] != tuple[0])
    {
      constant_function_arguments = false;
      break;
    }
  }
}




// *********************************************************




void FunctionArguments::initialise_almost_constant_function_arguments()
{
  acfa_major_value = 0;                              // Arbitrary
  acfa_minor_value = 0;
  acfa_minor_posn = 0;

  if (tuple.size() < 3)
  {
    almost_constant_function_arguments = false;
    return;
  }
  
  int first_val = tuple[0];
  nat first_val_size = 1;
  nat first_val_pos = 0;
  
  int second_val = 0;                 // Initialised to stop -Wall complaining
  nat second_val_size = 0;
  nat second_val_pos = 0;             // Initialised to stop -Wall complaining

  for (nat i = 1; i < tuple.size(); ++i)
  {
    if (tuple[i] == first_val)
    {
      ++first_val_size;
      continue;
    }
    
    if (second_val_size == 0)
    {
      second_val = tuple[i];
      second_val_size++;
      second_val_pos = i;
      continue;
    }
    
    if (tuple[i] == second_val)
    {
      ++second_val_size;
      continue;
    }
    
    almost_constant_function_arguments = false;
    return;
  }
  
  if (first_val_size <= 1)                    // Actually, 0 is impossible here
  {
    almost_constant_function_arguments = true;
    acfa_major_value = second_val;
    acfa_minor_value = first_val;
    acfa_minor_posn = first_val_pos;
    return;
  }
  
  if (second_val_size <= 1)
  {
    almost_constant_function_arguments = true;
    acfa_major_value = first_val;
    if (second_val_size == 0)                 // Constant tuple
    {
      acfa_minor_value = first_val;
      acfa_minor_posn = 0;
    }
    else                                      // Strictly almost-constant tuple
    {
      acfa_minor_value = second_val;
      acfa_minor_posn = second_val_pos;
    }
    return;
  }
  
  almost_constant_function_arguments = false;
  return;
}




// *********************************************************




bool FunctionArguments::is_almost_constant_function_arguments
                        (int& a, int& b, nat& bpos) const
{
  a = acfa_major_value;
  b = acfa_minor_value;
  bpos = acfa_minor_posn;
  return almost_constant_function_arguments;
}




// *********************************************************




void FunctionArguments::display(ostream &s) const
{
  display_tuple(s, tuple);
}




// *********************************************************




void FunctionArguments::display_perms() const
{
  display_tuple(cout, tuple);
  for (nat i = 0; i < get_number_of_permutations(); ++i)
    display_tuple(cout, get_permutation(i)->function_arguments_to_tuple());
}




// *********************************************************************
// FunctionArguments: Interfacing with the Function class
// *********************************************************************


nat FunctionArguments::max_functions_definable = 0;
set<nat> FunctionArguments::function_ids_available;




nat FunctionArguments::get_new_function_id()
{
  if (function_ids_available.size() == 0)
  {
    // No more new functions available.  Got to make room.
    // Arbitrarily, we create enough space for 10 more functions
    nat new_max_fns_def = max_functions_definable + 10;
    
    for (nat i = max_functions_definable; i < new_max_fns_def; ++i)
      function_ids_available.insert(i);
    
    for (map<vector<int>, const FunctionArguments*>::const_iterator
	   mi = tuple_to_fn_args_map.begin();
	 mi != tuple_to_fn_args_map.end();
	 ++mi)
    {
      (*mi).second->resize_functions_definable(new_max_fns_def);
    }
    max_functions_definable = new_max_fns_def;
  }

  // There's definitely enough room for a new function now
  nat new_id = *(function_ids_available.begin());
  function_ids_available.erase(new_id);

  return new_id;
}




// ***************




void FunctionArguments::destroy_function(nat id)
{
  if (id >= max_functions_definable)
    throw SanityError();
  
  function_ids_available.insert(id);
}








// *********************************************************************
// ***                     FunctionValues class                      ***
// *********************************************************************


map<set<int>, const FunctionValues*> FunctionValues::set_to_fn_values_map;
map<pair<const FunctionValues*, const FunctionValues*>,
  const FunctionValues*> FunctionValues::intersection_table;




void FunctionValues::FunctionValues_constructor(const set<int> &s)
{
  if (s.size() == 0)
    throw SanityError();

  values = s;
  size_of_set = values.size();
  min_val = min_set_value(values);
  max_val = max_set_value(values);
}




// *******************





void EmptyFunctionValues::FunctionValues_constructor(const set<int> &s)
{
  if (s.size() != 0)
    throw SanityError();
  values = s;
}




// *********************************************************************




const FunctionValues* FunctionValues::set_to_function_values(const set<int> &s)
{
  map<set<int>, const FunctionValues*>::const_iterator mi;
  mi = set_to_fn_values_map.find(s);
  if (mi != set_to_fn_values_map.end())
    return mi->second;
  
  FunctionValues* fv;
  if (s.size() == 0)
    fv = new(EmptyFunctionValues);
  else
    fv = new(FunctionValues);
  
  // Must store address *before* invoking "constructor"
  set_to_fn_values_map[s] = fv;
  fv->FunctionValues_constructor(s);
  
  return fv;
}




// *********************************************************************




// We only actually calculate intersections once each.  After that, we use
// a table lookup for speed.

const FunctionValues* FunctionValues::find_intersection
                      (const FunctionValues* fv1,
		       const FunctionValues* fv2)
{
  if (fv1 == fv2)
    return fv1;
  
  pair<const FunctionValues*, const FunctionValues*> fv_pair;
  
  if (fv1 > fv2)
    fv_pair = make_pair(fv1, fv2);
  else
    fv_pair = make_pair(fv2, fv1);

  // The first component of the pair is less than the second.  This makes
  // our intersection table half the size it would be otherwise.
  map<pair<const FunctionValues*, const FunctionValues*>, const FunctionValues*>::const_iterator mpi;
  mpi = intersection_table.find(fv_pair);
  if (mpi != intersection_table.end())
    return mpi->second;

  // The intersection's not in the table: we're going to have to calculate
  // it by hand.
  set<int> result;
  set_intersection((fv1->values).begin(),
		   (fv1->values).end(),
		   (fv2->values).begin(),
		   (fv2->values).end(),
		   inserter(result, result.begin()));
  
  return intersection_table[fv_pair] = set_to_function_values(result);
}







// *********************************************************************
// ***                       Function class                          ***
// *********************************************************************

pair<const FunctionArguments*, const FunctionValues*>
              Function::history_separator = make_pair(
                                (const FunctionArguments*) NULL,
                                (const FunctionValues*) NULL);




Function::Function(nat a,
		   const set<int>& dom,
		   const vector<const FunctionArguments*>& ord,
		   bool partial)
{
  function_id = FunctionArguments::get_new_function_id();
  arity = a;
  domain = dom;
  default_ordering = ord;
  is_partial = partial;
  history.clear();
  
  // Unless we're partial, the list of FunctionArgument pointers supplied
  // in ord ought always to be the |dom|^a distinct a-tuples over dom.
  // We don't fully sanity-check this, though the following two tests go
  // some way in this direction.

  if (!partial)
  {
    // In theory, this can overflow.  In practice, we hit the limits of
    // tractability long before the danger arises.  Anyway, the following
    // line of code is not critical, and can be safely removed
    if (dom.size() > 0 && (nat) pow((float) dom.size(), (int) a) != ord.size())
      throw SanityError();
    
    if (ord.size() > 0 && ord[0]->get_arity() != a)
      throw SanityError();
  }

  // Set up function.  Initially, all tuples from ord are given the set
  // of values dom Could use set_value(), but not using it is probably
  // neater, as it would put spurious entries on the history stack.
  const FunctionValues* fv = FunctionValues::set_to_function_values(dom);
  
  for (vector<const FunctionArguments*>::const_iterator favi = ord.begin();
       favi != ord.end();
       ++favi)
  {
    (*favi)->set_function_result(function_id, fv);
  }
}




// ***************




Function::~Function()
{
  FunctionArguments::destroy_function(function_id);
}




// ***************




Function* Function::clone() const
{
    Function* new_fn = new(Function)(arity,
				     domain,
				     default_ordering,
				     is_partial);
    nat new_fn_id = new_fn->function_id;
  
    // Copy the current function into the new function
    for (vector<const FunctionArguments*>::const_iterator favi =
	   default_ordering.begin();
         favi != default_ordering.end();
	 ++favi)
    {
      (*favi)->set_function_result(new_fn_id,
				   (*favi)->get_function_result(function_id));
    }
    
    return new_fn;
}




// ***************




void Function::display_as_list
                       (ostream &s,
			const vector<const FunctionArguments*>& ordering) const
{
  for (nat i = 0; i < ordering.size(); ++i)
  {
    s << "f(" << ordering[i] << ") = "
      << (*this)[ordering[i]] << "\n";
  }
}




// ****************




void Function::display_as_relation
                     (ostream &s,
		      const vector<const FunctionArguments*>& ordering) const
{
  bool first_tuple = true;
  set<int> noughtset;
  noughtset.insert(0);

  // Find a FunctionValues correcponding to {0}
  const FunctionValues* nought_fv =
    FunctionValues::set_to_function_values(noughtset);
  s << "{";
  for (nat i = 0; i < ordering.size(); ++i)
  {
    if ((*this)[ordering[i]] != nought_fv)
    {
      if (!first_tuple)
	s << ",";
      
      first_tuple = false;
      s << ordering[i];
    }
  s << "}\n";
  }
}




// ***************




void Function::restore_state()
{
  for(;;)
  {
    nat size = history.size();
    if (size == 0)
      throw Underflow();
    
    pair<const FunctionArguments*, const FunctionValues*> this_pair =
                                                               history[size-1];
    history.pop_back();
    
    if (this_pair == history_separator)
      break;
    
    // Can't use set_value in next line, because that changes the history.
    (this_pair.first)->set_function_result(function_id, this_pair.second);
  }
}



// ***************

