// **************************************************************************
//
// relation.cc
//
// Part of Polyanna, Richard Gault, 2001
//
// Copyright (C) University of Oxford, 2001, 2002
//
// **************************************************************************



// ******************************************
//    Implementation of Relation class
// ******************************************

#include "relation_headers.h"
#include "function_headers.h"
#include "misc_headers.h"
#include <iostream>
#include <cmath>

using namespace std;



Relation::Relation(nat a)
{
  arity = a;
  redundant = false;
  type = Unknown;
  top_level = true;
  squashed = false;
}



// ******************************************




void Relation::change_arity(nat a)
{
  if (arity != a && get_size() != 0)
    throw AritiesInconsistent();

  arity = a;
}



// ******************************************




set<int> Relation::get_apparent_domain() const
{
  set<int> d;
  if (d.size() != 0)
  {
    cerr << "ERROR: Newly created set has non-zero size\n";
    throw SanityError();
  }

  for (RI tp = set_of_tuples.begin(); tp != set_of_tuples.end(); tp++)
    for (nat i = 0; i < arity; i++)
      d.insert((*tp)[i]);

  return d;
}



// ******************************************




void Relation::set_real_domain(const set<int> &s)
{
  if (!is_subset(get_apparent_domain(), s))
    throw DomainError();
  
  // Set the real domain, and propagate to our children
  real_domain = s;
  for (set<Relation*>::iterator rpi = children.begin();
       rpi != children.end();
       ++rpi)
    (*rpi)->set_real_domain(s);
}



// ******************************************




bool Relation::add_tuple(const vector <int> &t)
{
  if (type != Unknown)
    throw BelatedAlteration();
  
  change_arity(t.size());    // Will throw AritiesInconsistent if appropriate
  
  return (set_of_tuples.insert(t)).second;
}



// ******************************************




bool Relation::remove_tuple(const vector<int> &t)
{
  if (type != Unknown)
    throw BelatedAlteration();
  
  if (t.size() != arity)
    throw AritiesInconsistent();

  nat n = set_of_tuples.erase(t);
  if (n == 0) return false;
  if (n == 1) return true;

  // We've got two copies of a tuple in our set.  Shouldn't happen.
  cerr << "ERROR: A set contains two identical elements\n";
  throw SanityError();
}



// ******************************************




void Relation::squash_with_map(const map<int,int>& squash_map)
{
  if (squashed)
  {
    cerr << "Warning: A relation has been squashed twice.  This is not illegal,\n";
    cerr << "         but may indicate a programming error.\n";
  }
  
  // Apply the squashing map to ourself.  We have to create a new set of
  // tuples, because we can't modify the old one in situ.
  set < vector<int> > new_set_of_tuples;
  new_set_of_tuples.clear();
  vector<int> new_tuple(get_arity());

  for (RI tp = set_of_tuples.begin(); tp != set_of_tuples.end(); tp++)
  {
    const vector<int>& old_tuple = *tp;
    
    for (nat i = 0; i < get_arity(); ++i)
    {
      map<int,int>::const_iterator mi = squash_map.find(old_tuple[i]);
      
      if (mi == squash_map.end())
	throw SanityError();
      
      new_tuple[i] = mi->second;
    }
    new_set_of_tuples.insert(new_tuple);
  }
  
  set_of_tuples.swap(new_set_of_tuples);

  // Apply the squashing map to our children (if any)
  for (set<Relation*>::const_iterator rpi = children.begin();
       rpi != children.end();
       ++rpi)
    (*rpi)->squash_with_map(squash_map);
}





// ************************************************************************
// ***           Functions which perform decompositions                 ***
// ************************************************************************




// This is the public interface to the decomposition functions.  It tries
// all decompositions in turn, stopping when it finds one which works.
// The order of testing is important in some cases.

bool Relation::decompose()
{
  if (apply_decomposition(&Relation::check_for_trivial_relations))
    return false;
  if (apply_decomposition(&Relation::check_for_duplicated_columns))
    return true;
  if (apply_decomposition(&Relation::check_for_allows_anything))
    return true;
  if (apply_decomposition(&Relation::check_for_cartesian_power))
    return true;
  if (apply_decomposition(&Relation::check_for_cartesian_product))
    return true;
  if (apply_decomposition(&Relation::check_for_disjoint_domains))
    return true;
  if (apply_decomposition(&Relation::check_for_conjunctive_product))
    return true;
  
  // and so on...
  
  type = OtherIrreducible;   // TrivialIrreducible has already been dealt with
  
  return false;
}




// *******************************************************************




// Apply a decomposition, then recursively decompose any children
// which have been generated.

bool Relation::apply_decomposition(DecompositionFunction decomposer)
{
  bool result = (this->*decomposer)();
  
  if (!result)
    return false;

  // A successful decomposition took place.  Some sanity checks.
  if (type == Unknown)
    throw DecompositionError();
  
  if (!get_redundant() && type != DisjointDomains && type != ConjunctiveProduct)
    throw DecompositionError();

  // Decompose children.
  for (set<Relation*>::const_iterator rpi = children.begin();
       rpi != children.end();
       ++rpi)
  {
    // Shouldn't (normally) have sorted out children yet
    if ((*rpi)->type != Unknown)
      throw DecompositionError();

    // Nor should children have been determined redundant
    if ((*rpi)->get_redundant())
      throw DecompositionError();
    
    (*rpi)->decompose();
  }
  return true;
}




// *********************************************************************




Relation* Relation::make_top_level_copy()
{
  Relation* copy = new(Relation)(arity);
  
  copy->set_of_tuples = set_of_tuples;
  copy->set_real_domain(real_domain);
  
  return copy;
}




// **********************************************************************




Relation* Relation::make_child(Relation* copy_from)
{
  Relation* new_child = new(Relation);
  children.insert(new_child);
  new_child->set_real_domain(get_real_domain());
  new_child->top_level = false;
  
  if (copy_from != NULL)
  {
    new_child->change_arity(copy_from->get_arity());
    new_child->set_of_tuples = copy_from->set_of_tuples;
  }
  
  if (!is_subset(new_child->get_apparent_domain(),
		 new_child->get_real_domain()))
    throw DomainError();
  
  return new_child;
}




// **********************************************************************




bool Relation::check_for_trivial_relations()
{
   if (get_arity() == 0 || get_size() == 0)
   {
     redundant = true;
     type = TrivialIrreducible;
     return true;
   }
   
   return false;
}




// ***********************************************************************




// The algorithm works by putting each column in turn into a set.  If
// the column is already present, keep[that_column] is set to true.
// Finally, delete_columns() is called to shrink the relation.

bool Relation::check_for_duplicated_columns()
{
  nat size = get_size();
  vector <bool> keep (get_arity());
  set< vector<int> > set_of_columns;
  bool deleted_columns = false;
  
  // Set up the vector keep of columns we're going to keep.
  for (nat i = 0; i < get_arity(); i++)
  {
    vector<int> col (size);   // Just a temp var to hold column i
    nat j = 0;
    
    for (RI tp = set_of_tuples.begin(); tp != set_of_tuples.end(); tp++, j++)
      col[j] = (*tp)[i];
    
    keep[i] = (set_of_columns.insert(col)).second;
    deleted_columns = deleted_columns || !(keep[i]);
  }
  if (!deleted_columns)
    return false;

  // Actually delete the columns
  type = DuplicatedColumns;
  redundant = true;
  Relation* child = make_child(this);
  child->delete_columns(keep);
  
  return true;
}




// ************************************************************************




nat Relation::delete_columns(const vector<bool>& keep)
{
  if (keep.size() != arity)
    throw AritiesInconsistent();
  
  nat new_arity = 0;
  
  for (nat i = 0; i < arity; i++)
    if (keep[i]) new_arity++;
  
  if (new_arity == arity)
    return 0;

  // Must shrink relation.  So we copy the shortened tuples to a new set
  set< vector<int> > new_set_of_tuples;
  vector<int> t(new_arity);
  nat from, to;
  for (RI tp = set_of_tuples.begin(); tp != set_of_tuples.end(); tp++)
  {
    for (to = 0, from = 0; from < arity; to++, from++)
    {
      // Find column to copy from
      while (from < arity && keep[from] == false)
	from++;
      
      if (from >= arity)
	continue;    // Done copying this tuple
      
      // Copy the tuple element
      t[to] = (*tp)[from];
    }
    new_set_of_tuples.insert(t);
  }
  
  // Set the present relation to have the new set of tuples
  set_of_tuples.swap(new_set_of_tuples);
  nat number_deleted = arity-new_arity;
  arity = new_arity;
  
  return number_deleted;
}




// *****************************************************************




bool Relation::check_for_allows_anything()
{
  if (arity != 1)
    return false;
  
  if (get_size() != get_real_domain_size())
    return false;

  // Success
  type = AllowsAnything;
  redundant = true;
  Relation* child = make_child(NULL);   // Create child as empty relation
  
  if (child->get_size() != 0)
  {
    cerr << "ERROR: Child relation of non-zero size created\n";
    throw SanityError();
  }
  
  return true;
}




// *****************************************************************




bool Relation::check_for_cartesian_power()
{
  nat relation_domain_size = get_apparent_domain_size();
  nat relation_size = get_size();

  // Trivial relations should already have been picked up as being
  // TrivialIrreducible
  if (relation_size == 0 || relation_domain_size == 0 || arity == 0)
    throw SanityError();

  if (arity == 1)
    return false;                    // can't improve on this

  // There's a concern with overflow on the following line, which is why
  // we have all the mucking about with casts to double.
  if ( pow( (double) relation_domain_size, (int) arity) > (double) relation_size)
    return false;
  
  if ( pow( (double) relation_domain_size, (int) arity) < (double) relation_size)
    throw SanityError();

  // We now know that the relation is a Cartesian power.
  // Reduce it to an arity-1 relation.

  vector <bool> keep (arity);
  
  for (nat i = 1; i < arity; i++)
    keep[i] = false;                  // Better safe than sorry  
  keep[0] = true;

  type = CartesianPower;
  redundant = true;
  Relation* child = make_child(this);
  child->delete_columns(keep);
  
  return true;
}




// ******************************************************************




bool Relation::check_for_cartesian_product()
{
  nat relation_size = get_size();
  
  if (arity <= 1) return false;
  if (relation_size == 0) return false;
  if (relation_size == 1)
  {
    // Split the size-1, arity-a relation up into 'a' arity-1 relations.
    Relation* child;
    vector <int> t (1);
    
    for (nat i = 0; i < arity; i++)
    {
      type = CartesianProduct;
      redundant = true;
      child = make_child(NULL);
      t[0] = (*set_of_tuples.begin())[i];
      child->add_tuple(t);
    }
    return true;
  }

  // We are now guaranteed to have size >= 2 and arity >= 2.
  // Try splitting the relation at the (arity-1) natural points.
    vector <bool> split (arity);
    
    for (nat i = 0; i < get_arity(); i++)
      split[i] = false;
    
    for (nat i = 0; i < get_arity() - 1; i++)
    {
      split[i] = true;
      if (check_split_for_cartesian_product(split))
	return true;
    }

  // Next method:
  // For each column, find HCF of number of occurrences of each domain
  // element.
  typedef multiset<int>::const_iterator MI;
  vector <nat> hcf(arity);
  
  for (nat col = 0; col < arity; col++)
  {
    multiset<int> dom_elements;
    
    for (RI ti = set_of_tuples.begin(); ti != set_of_tuples.end(); ++ti)
      dom_elements.insert((*ti)[col]);
    
    pair <MI, MI> ran;
    set <nat> occurrences;
    
    for (MI di = dom_elements.begin();
	 di != dom_elements.end();
	 di = ran.second)
    {
      ran = dom_elements.equal_range(*di);
      occurrences.insert(distance(ran.first,ran.second));
    }
    
    // (Now just need to find the HCF of all the values in occurrences.)
    if (occurrences.size() == 0)
      throw SanityError();

    nat this_hcf = *(occurrences.begin());
    
    for (set<nat>::const_iterator ii = occurrences.begin();
	 ii != occurrences.end() && this_hcf != 1;
	 ++ii)
      this_hcf = highest_common_factor(this_hcf, *ii);
    
    hcf[col] = this_hcf;
  }
  
  // Search for columns with an HCF equal to the size of the relation.
  // These must be constant columns, and may be cast off.
  vector<bool> keep (arity);
  bool found_constant_column = false;
  
  for (nat i = 0; i < arity; i++)
  {
    if (hcf[i] == relation_size)
    {
      found_constant_column = true;
      keep[i] = false;
      vector <int> tuple (1);
      tuple[0] = (*(set_of_tuples.begin()))[i];
      Relation* child = make_child(NULL);
      child->add_tuple(tuple);
    }
    else
      keep[i] = true;
  }
  
  if (found_constant_column)
  {
    type = CartesianProduct;
    redundant = true;
    Relation* child = make_child(this);
    child->delete_columns(keep);
    return true;
  }

  // There are no constant columns, so if there are any columns with an
  // HCF of 1, then the relation is not a Cartesian product.  Return
  // false.
  for (nat i = 0; i < arity; i++)
    if (hcf[i] == 1)
      return false;

  // This is the tricky case.  It seems likely that the relation is a
  // Cartesian product.  As a heuristic, we try splitting the relation
  // up as follows.  Any columns whose HCF-value has an HCF of more
  // than 1 with the HCF-value of the first column, are put in the
  // first relation.  Any other columns are put in the second
  // relation.  In most (but not all) cases, this will be the correct
  // split (if there is one).  One common case where this heurisitc
  // fails is if the relation is the Cartesian power case.  But that's
  // already dealt with.

  for (nat i = 0; i < arity; i++)
    if (highest_common_factor(hcf[0], hcf[i]) > 1)
      split[i] = true;       // NB: split[0] will always be set to true by this
    else
      split[i] = false;
  
  if (check_split_for_cartesian_product(split))
    return true;

  // Provided the arity is not too large, we can try splitting the
  // relation up by trying every possibility with brute force.
  if (arity <= 10)
  {
    split[0] = false;
    nat max_pattern = (nat) pow((float) 2 , (int) arity - 1) - 1;
    
    for (nat pattern = 1; pattern <= max_pattern; pattern++)
    {
      nat pattern_copy = pattern;
      for (nat i = 1; i < arity; i++)
      {
        split[i] = pattern_copy % 2;
	pattern_copy /= 2;
      }
      if (check_split_for_cartesian_product(split))
	return true;
    }
    return false;
  }

  // Finally, warn that our decomposition efforts have failed.
  cerr << "Warning: A relation was found which may be the Cartesian product\n";
  cerr << "         of two other relations, but it has too great an arity for\n";
  cerr << "         an actual decomposition to be found (its arity is " << arity << ", the\n";
  cerr << "         maximum decomposable arity is " << 10 << ").  The relation has been\n";
  cerr << "         left un-decomposed.\n";
  
  return false;
}



// ********************************************************




bool Relation::check_split_for_cartesian_product(const vector<bool>& split)
{
  if (split.size() != get_arity())
    throw AritiesInconsistent();
  
  // First generate the two projected relations
  vector <bool> inverse_split (split.size());
  
  for (nat i = 0; i < arity; i++)
    inverse_split[i] = !split[i];
  
  Relation first = *this;
  Relation second = *this;
  first.delete_columns(split);
  second.delete_columns(inverse_split);

  if (first.arity == 0 || second.arity == 0)
    return false;

  // Is the produce of their sizes equal to the size of the present
  // relation?  Yes ==> genuine split, No ==> not a split.
  if (first.get_size() * second.get_size() != get_size())
    return false;

  // We have a genuine split!  Make permanent copies of first and second,
  // and make them our children.
  make_child(&first);
  make_child(&second);
  type = CartesianProduct;
  redundant = true;
  
  return true;

}




// **********************************************************************




bool Relation::check_for_disjoint_domains()
{
  // Disjoint domains are unlikely to be useful in the arity-1 case
  if (arity < 2) return false;

  // We put the tuples of the relation into numbered equivalence classes.
  // Two tuples which overlap must share the same number.
  nat relation_size = get_size();
  
  // eq_class should really be a map indexed by tuples.  But this is faster,
  // and the variable is local to the current function.
  vector <nat> eq_class (relation_size);
  RI tp, t2p, t3p;
  nat tp_number, t2p_number, t3p_number;

  for (tp = set_of_tuples.begin(), tp_number = 0;
       tp != set_of_tuples.end();
       ++tp, ++tp_number)
  {
    eq_class[tp_number] = tp_number;     // Assume tuple is not connected to anything
    for (t2p = set_of_tuples.begin(), t2p_number = 0;
	 t2p != tp;
	 ++t2p, ++t2p_number)
    {
      if (eq_class[t2p_number] == eq_class[tp_number])
	continue;
      
      // (Check if tuples tp and t2p have any elements in common.
      // This is currently quadratic, but only in the arity of
      // the relation, so it's probably not worth changing.)
      for (nat i = 0; i < arity; i++)
	for (nat j = 0; j < arity; j++)
	  if ((*tp)[i] == (*t2p)[j])
	    goto out_of_loop;            // You didn't see this
      continue;
      
    out_of_loop:
      nat min_class_val = (eq_class[t2p_number] < eq_class[tp_number]) ?
	eq_class[t2p_number] :
	eq_class[tp_number];
      
      nat max_class_val = (eq_class[t2p_number] > eq_class[tp_number]) ?
	eq_class[t2p_number] :
	eq_class[tp_number];
      
      eq_class[t2p_number] = eq_class[tp_number] = min_class_val;
      
      // (Any tuple whose equivalence number was max_class_val
      // now needs updating.)
      for (t3p = set_of_tuples.begin(), t3p_number = 0;
	   t3p != tp;
	   ++t3p, t3p_number++)
	if (eq_class[t3p_number] == max_class_val)
	  eq_class[t3p_number] = min_class_val;	
    }
  }

  // eq_class is now set up.
  // We now do the actual splitting, beginning by caculating the
  // number of equivalence classes, and their maximum value (which
  // will always be <= size of the original relation.)

  set<nat> distinct_classes;
  nat max_class_number = 0;
  
  for (tp = set_of_tuples.begin(), tp_number = 0;
       tp != set_of_tuples.end();
       ++tp, ++tp_number)
  {
    distinct_classes.insert(eq_class[tp_number]);
    
    if (eq_class[tp_number] > max_class_number)
      max_class_number = eq_class[tp_number];
  }

  if (distinct_classes.size() == 1)
    return false;

  // (We now know that we have genuine disjoint domains.)
  vector <Relation*> sub_relations (max_class_number + 1);
  
  for (set<nat>::const_iterator cvi = distinct_classes.begin();
       cvi != distinct_classes.end();
       ++cvi)
    sub_relations[*cvi] = new(Relation);
  
  for (tp = set_of_tuples.begin(), tp_number = 0;
       tp != set_of_tuples.end();
       ++tp, ++tp_number)
    sub_relations[eq_class[tp_number]]->add_tuple(*tp);
  
  for (set<nat>::const_iterator cvi = distinct_classes.begin();
       cvi != distinct_classes.end();
       ++cvi)
    make_child(sub_relations[*cvi]);
  
  type = DisjointDomains;              // Not redundant
  return true;
}




// *********************************************************************




bool Relation::check_for_conjunctive_product()
{
  return false;
}




// *********************************************************************




void Relation::construct_constraints(ConstraintStore& cs) const
{
  nat fn_arity = cs.get_arity();
  nat rln_arity = get_arity();                  // For clarity
  vector<RI> tuple_of_tuples(fn_arity);

  // Check for trivial cases
  if (rln_arity == 0 || set_of_tuples.size() == 0)
    return;

  // Generate each fn_arity-tuple of tuples from this relation in turn,
  // and form a constraint from it
  for (nat i = 0; i < fn_arity; ++i)
    tuple_of_tuples[i] = set_of_tuples.begin();

  for (;;)
  {
    nat i;               // Need these variables pre-defined because we can't
    bool reset = true;   // declare them all in the for-loop initialiser below
    int j;
    
    // (Form a constraint)
    // (First generate, as a vector of length rln_arity, the
    // FunctionArguments we're constraining)
    vector <const FunctionArguments*> constrained_fav(rln_arity);
    
    for (i = 0; i < rln_arity; ++i)
    {
      vector<int> this_tuple(fn_arity);
      
      for (nat j = 0; j < fn_arity; ++j)
	this_tuple[j] = (*(tuple_of_tuples[j]))[i];
      
      constrained_fav[i] = FunctionArguments::tuple_to_function_arguments(this_tuple);
    }
    
    // (Convert the vector into a set (to remove duplicated FAs))
    set <const FunctionArguments*> constrained_fas;
    copy(constrained_fav.begin(), constrained_fav.end(),
	 inserter(constrained_fas, constrained_fas.end()));
    
    // (Generate the new constraint)
    PolymorphismConstraint* pc =  cs.add_polymorphism_constraint(constrained_fas);

    // (Add each possible simultaneous assignment of values to the
    // FunctionArguments to the constraint)
    // Don't worry about any inconsistencies -- eg. f(0,0) == 1 && f(0,0) == 0.
    // It's up to the constraint itself to deal with these.
    for (RI rpi = set_of_tuples.begin(); rpi != set_of_tuples.end(); ++rpi)
    {
      set<pair<const FunctionArguments*, int> > disjunct;
      
      for (i = 0; i < rln_arity; ++i)
	disjunct.insert(make_pair(constrained_fav[i], (*rpi)[i]));
      
      pc->add_possible_assignments(disjunct);
    }
    
    // (Find lexicographically next tuple-of-tuples)
    for (j = fn_arity - 1, reset = true; j >= 0 && reset; --j)    // Can't use nat, as j goes briefly -ve
    {
      ++(tuple_of_tuples[j]);
      
      if (tuple_of_tuples[j] == set_of_tuples.end())
	tuple_of_tuples[j] = set_of_tuples.begin();
      else
	reset = false;
    }
    
    if (reset) break;     // We've "clocked over" to the first tuple of tuples once again
  }
}




// **********************************************************************




void Relation::hierarchical_output(ostream& s) const
{
  switch(type)
  {
  case Unknown:
  {
    s << "Undetermined relation: ";
    output_as_tuples(s, false);
    return;
  }
  case TrivialIrreducible:
  {
    s << "Trivial relation ";
    output_as_tuples(s, false);
    return;
  }
  case OtherIrreducible:
  {
    s << "Irreducible relation ";
    output_as_tuples(s, false);
    return;
  }
  case DuplicatedColumns:
  {
    s << "AddColumns(";
    break;
  }
  case CartesianPower:
  {
    s << "CartesianPower(";
    break;
  }
  case CartesianProduct:
  {
    s << "CartesianProduct(";
    break;
  }
  case DisjointDomains:
  {
    s << "DisjointUnion(";
    break;
  }
  case ConjunctiveProduct:
  {
    s << "ConjunctiveProduct(";
    break;
  }
  case AllowsAnything:
  {
    s << "AllowsAnythingExpansion(";
    break;
  }
  default:
  {
    throw SanityError();
  }
  }
  
  for (set<Relation*>::const_iterator rpi = children.begin();
       rpi != children.end();
       // Null statement
       )
  {
    (*rpi)->hierarchical_output(s);
    
    if (++rpi != children.end())
      s << ",";
  }
  s <<")";
}




// **********************************************************************




void Relation::output_as_tuples(ostream& s, bool new_lines) const
{
  s << "{";
  
  for (RI tp = set_of_tuples.begin(); tp != set_of_tuples.end(); ++tp)
  {
    if (tp != set_of_tuples.begin())
    {
      if (new_lines) s << ",\n "; else s << ", ";
    }
    display_tuple(s, *tp);
  }
  s << "}";
}




// **********************************************************************



// ***********************************************************************
//   The following functions are friends of the Relation class
// ***********************************************************************


istream& operator>> (istream& s, Relation& r)
{
  if (r.arity < 0) throw Relation::SanityError();
  char c;

  // First character of relation: should be '{'
  if (s >> c, c != '{') throw Relation::ReadError();
  
  while (s >> c, c != '}')
  {
    if (c == ',') s >> c;        // Optional comma before/after each tuple
    if (c == '}') break;
    if (c != '<') throw Relation::ReadError();
    
    vector<int> v(0);            // We'll grow v to fit the tuple we're reading
    
    if (s >> c, c != '>')        // Provided our tuple is of length at least 1
    {
      s.putback(c);
      
      while (c != '>')
      {
	if (s.eof()) throw Relation::ReadError();     // Unexpected EOF

	s >> c; s.putback(c);
	// May as well allow negative ints!
	if (!((c >= '0' && c <= '9') || c == '-'))
	  throw Relation::ReadError();
	
	int next_value;
	s >> next_value;
	v.push_back(next_value);
	s >> c;
	
	if (c != ',' && c != '>') throw Relation::ReadError();
      }
    }
    
    // We've now read in a whole tuple.  Let's add it to the relation.
    r.add_tuple(v);   // Will throw AritiesInconsistent if appropriate
  }
  return s;
}
