// *******************************************************************
//
// symmetry_headers.h
//
// Part of Polyanna, Richard Gault, 2001
//
// Copyright (C) University of Oxford, 2001, 2002
//
// *******************************************************************

#ifndef SYMMETRY_HEADERS
#define SYMMETRY_HEADERS

#include "constraints_headers.h"

class RelationStore;

class AutomorphismConstraint: public Constraint
{
 public:
  // Constructor.  Supply FAs we affect, along with their domain, and the
  // RelationStore whose polymorphisms we're searching for, and four bools.
  // The bools tell us which automorphisms to enable: argument permutations,
  // automorphisms applied to the left, to the right, or to both sides of a
  // function.  The default bool indicates whether we want accuracy, or whether
  // (if false) we would prefer the constraint to be faster, but not as good
  // at weeding out essentially identical functions.  (When false, the
  // functions we prohibit are a subset of those we prohibit when true.)
  AutomorphismConstraint(const std::set<const FunctionArguments*> &fas,
			 const std::set<int>& dom,
			 const RelationStore& relstore,
			 bool argument_perms,
			 bool da_left,
			 bool da_right,
			 bool da_both,
			 bool accuracy = true);

  // The following functions are all derived from the Constraint class.
  // See that class's declaration for the meaning of the functions.
  
  ConstraintType get_constraint_type() const
    {return Automorphism;}

  bool permutation_closed() const
    {return true;}

  bool need_level_notification() const
    {return false;}                       // May be true one day
  
  Constrainedness is_constraint_relevant(const FunctionArguments* fa) const
    {return SymmetryConstrained;}

  SatLevel apply_constraint
            (Function &f,
	     const std::set<const FunctionArguments*> &modified_fas,
	     std::set<const FunctionArguments*> &new_modified_fas,
	     bool complete) const;

  void display(std::ostream& s) const;

  // The following functions are not inherited from the Constraint class

  // Which of the four sub-constraints are currently enabled?
  bool argument_permutations_enabled() const
    {return argument_perms;}
  
  bool domain_automorphism_left_enabled() const
    {return da_left;}
  
  bool domain_automorphism_right_enabled() const
    {return da_right;}
  
  bool domain_automorphism_both_enabled() const
    {return da_both;}

 private:
  
  // The domain we're working over
  std::set<int> domain;

  // Are argument perms enabled?
  bool argument_perms;

  // Or domain automorphisms applied on left...
  bool da_left;

  // ...on right...
  bool da_right;

   // ...or on both sides?
  bool da_both;

  // Prefer accuracy over speed?
  bool accuracy;
  
  // Typedef of vector of FunctionArguments.  Just to save on typing.
  typedef std::vector<const FunctionArguments*> FA_VEC;

  // Typedef of an automorphism of the domain
  typedef std::map<int, int> DOMAIN_AUTOMORPHISM;

  // Describes generation of new fn from old
  typedef std::pair<FA_VEC, DOMAIN_AUTOMORPHISM> FN_GENERATOR;

  // Equivalence class generator
  typedef std::set<FN_GENERATOR> ECG;
  
  // Have we set up the equivalence class?
  mutable bool done_ecg_initialisation;

  // The order we put FAs in to check for minimality
  mutable FA_VEC fa_ordering;

  // The equivalence class generator we use
  mutable ECG ecg;

  // Set up the equivalence class generator
  void initialise_ecg(const std::set<const FunctionArguments*> &fas,
		      const RelationStore& relstore) const;

  // Set result = a1 . a2  (applying on left: result(x) = a1(a2(x)) )
  // Could be static, but isn't: because if it isn't then we can check
  // the domains of a1 & a2 are correct
  void compose_automorphisms(const DOMAIN_AUTOMORPHISM& a1,
			     const DOMAIN_AUTOMORPHISM& a2,
			     DOMAIN_AUTOMORPHISM& result) const;

  // Apply the automorphism pointwise to each FA in the vector.  Again,
  // could be static, but isn't
  void apply_automorphism_to_fa_vec(const DOMAIN_AUTOMORPHISM& da,
				    const FA_VEC& fa_vec,
				    FA_VEC& result) const;
  
  // Each of these functions (one for each of the four types of automorphism
  // symmetry we deal with) takes a set of function generators whcih always
  // generate functions in the same equivalence class, (as well, perhaps, as
  // some other pertinent information) and returns a set of further function
  // generators which always generate functions in the same equivalence class
  // as the original set of generators did.  With me?
  void add_fn_gens_argument_perms(const ECG fn_gens,
				  ECG& additional_fn_gens) const;
  
  void add_fn_gens_da_left(const ECG& fn_gens,
			   const std::set<DOMAIN_AUTOMORPHISM>& iufs,
			   ECG& additional_fn_gens) const;
  
  void add_fn_gens_da_right(const ECG& fn_gens,
			    const std::set<DOMAIN_AUTOMORPHISM>& iufs,
			    ECG& additional_fn_gens) const;
  
  void add_fn_gens_da_both(const ECG& fn_gens,
			   const std::set<DOMAIN_AUTOMORPHISM>& iufs,
			   ECG& additional_fn_gens) const;
};






// ********************************************************************




// An EssentiallyNAryConstraint ensures that each function found uses
// all its arguments.  This is the only symmetry-breaking constraint which
// is not related to the AutomorphismsConstraint

class EssentiallyNAryConstraint : public Constraint
{
 public:
  // Constructor.  Supply set of those FunctionArguments which define the
  // arguments of functionms we're searching for, along with the domain
  // and arity of the functions we're searching for.
  // (The first of these could be deduced from the other two, but we
  // pass all three to save ourselves from having to calculate the set
  // twice.)
  EssentiallyNAryConstraint(const std::set<const FunctionArguments*>& fas,
			    const std::set<int>& dom,
			    nat a);

  ConstraintType get_constraint_type() const
    {return EssentiallyNAry;}

  bool need_level_notification() const
    {return true;}

  bool permutation_closed() const
    {return true;}

  Constrainedness is_constraint_relevant(const FunctionArguments* fa) const
    {return SymmetryConstrained;}

  void check_consistency(nat a,
			 const std::set<int>& dom,
			 const std::set<const FunctionArguments*>& fas) const;

  SatLevel apply_constraint
            (Function &f,
	     const std::set<const FunctionArguments*> &modified_fas,
	     std::set<const FunctionArguments*> &new_modified_fas,
	     bool complete) const;


  // About to start search.  Max search depth supplied to us as an argument
  void notify_search_starting(nat max_d);

  // Moving down a search level
  void notify_search_descent();

  // Moving up a search level
  void notify_search_ascent();

  // The search ahs finished
  void notify_search_ending();    

  void display(std::ostream& s) const;

 private:
  // modifications[fa][i] contains a list of FAs which are obtained from
  // fa by modifying its i'th element (or, more accurately, the i'th element
  // of its corresponding tuple).
  std::map<const FunctionArguments*,
         std::vector<std::list<const FunctionArguments*> > > modifications;
  
  // A list of function positions which we haven't yet determined are essential
  mutable std::list<nat> undetermined_fn_posns;
  
  // For each search depth, a list of function positions which we found
  // out were essential at that depth
  mutable std::vector<std::list<nat> > essential_fn_posns;

  // Arity of polymorphism we're searching for
  nat arity;

  // The domain we're working over
  std::set<int> domain;

  // Current search depth
  nat depth;

  // Maximum search depth
  nat max_depth;
};

#endif
